//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
// $Id: ctrlc.cxx,v 1.1 2000/06/21 08:10:19 loreti Exp $
//
// Control-C handler.  When called the first time, ctrlc()
// establishes the handler; successive calls return "true" if an
// interrupt has been detected, and "false" otherwise.
//
// -------------------------------------------------------------

#include <iostream>
#include <cstdio>
extern "C" {
  #include <sys/types.h>
  #include <unistd.h>
  #include <signal.h>
}
#include "sol.hh"

using std::cerr;
using std::endl;

enum status { INIT, RUN, INTERRUPT };

static status currentStatus(INIT);
static void sigHandler(const int);

bool ctrlc()
{
  struct sigaction sa;
  sigset_t         oldMask;

  switch (currentStatus) {
    case INIT:

// Initial call.  We first block all the signals, except:
// - SIGKILL (associated with "kill -9 <PID>") and SIGSTOP (that
//    suspends the process): by definition SIGKILL and SIGSTOP
//    cannot be ignored;
// - SIGCONT (that resumes a stopped process);
// - SIGTERM (used to terminate the program via a "kill <PID>"
//   command from the shell;
// - SIGINT (the interrupt signal from a Control-C).
// When done, we associate the local function "sigHandler" to
// the SIGTERM and SIGINT signals; the initial call returns
// "true" in case of success, of "false" if an error occurred.

      sigfillset(&sa.sa_mask);
      sigdelset(&sa.sa_mask, SIGKILL);
      sigdelset(&sa.sa_mask, SIGSTOP);
      sigdelset(&sa.sa_mask, SIGCONT);
      sigdelset(&sa.sa_mask, SIGTERM);
      sigdelset(&sa.sa_mask, SIGINT);
      if (sigprocmask(SIG_SETMASK, &sa.sa_mask, &oldMask) != 0) {
        cerr << "Couldn't block signals\n";
        perror("sigprocmask");
        return false;
      }

      sa.sa_handler = sigHandler;
      sa.sa_flags = 0;
      if (sigaction(SIGTERM, &sa, 0)   ||   sigaction(SIGINT, &sa, 0)) {
        cerr << "Couldn't set a signal handler\n";
        perror("sigaction");
        return false;
      }

      currentStatus = RUN;
      return true;

    case RUN:
      return false;

    case INTERRUPT:
      currentStatus = RUN;
      return true;

    default:
      cerr << "Can't happen: " << __FILE__ << ", line "
           << __LINE__ << endl;
      return true;
  }
}

static void sigHandler(
  const int sigNum
) {
  switch (sigNum) {
    case SIGTERM:
    case SIGINT:
      currentStatus = INTERRUPT;
      break;

    default:
      cerr << "Can't happen: " << __FILE__ << ", line "
           << __LINE__ << endl;
  }
}
