//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
// $Id: eventloop.cxx,v 1.1 2000/06/21 08:10:19 loreti Exp $
//
// The Montecarlo main loop.  Here, for the needed number of
// times:
// - the pack is shuffled;
// - a card table is created from the shuffled pack (that is not
//   changed --- in particular cases you may want to print the
//   pack that corresponds to a particular situation);
// - we try to solve the patience, calling the "solve" method
//   for the table;
// - a summary of the results is updated, and the Control-C
//   handler is checked;
// - at regular intervals, the number of tries is printed to
//   tell to the user that something is happening :-) ;
// - the summaries are also written on file.
//
// -------------------------------------------------------------

#include <iostream>
#include <fstream>
#include <cstring>
#include "sol.hh"
#include "table.hh"

using namespace std;

// Local functions

static void print(unsigned long, const unsigned int[], ostream &);

void eventLoop(
  const unsigned long maxTries,
  const unsigned long nStep
) {
  pack p;
  unsigned long nTries;
  unsigned int hits[card::totCards+1];

  memset(hits, 0, sizeof(hits));
  ofstream os(outFileName);

#ifdef DEBUG
  cout << "*** Initial pack:\n" << p << endl;
#endif

  for (nTries = 0;   nTries < maxTries;   nTries++) {
    if (ctrlc()) break;
    if (nStep   &&   (nTries % nStep) == 0) {
      cout << '\r' << nTries;
      cout.flush();
#ifdef DEBUG
      if (nTries) print(nTries, hits, os);
#endif
    }

    p.shuffle();
#ifdef DEBUG
    cout << "*** Trial " << nTries << " - shuffled pack:\n"
         << p << endl;
#endif

    table t(p);
#ifdef DEBUG
    cout << t << endl;
#endif

    unsigned int res = t.solve();
#ifdef DEBUG
    cout << "*** Returned: " << res << endl << endl;
#endif

    (hits[res])++;
  }

  if (nStep) cout << '\r' << maxTries << endl;
  print(nTries, hits, cout);
  print(nTries, hits, os);
}

static void print(
  unsigned long n,
  const unsigned int hits[],
  ostream &os
) {
  if (os) {
    os << "*** Hits/" << n << " : " << hits[0] << "\n    ";
    for (unsigned long i = 1;   i <= card::totCards;   ) {
      os << i << ':' << hits[i];
      if (++i % 5  ==  1) os << "\n    "; else os << ' ';
    }
    os << endl;
  }
}
