//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
// $Id: sol.cxx,v 1.3 2006/09/11 12:28:26 loreti Exp $
//
// This is a main program written in order to analyze a card
// patience game using a Montecarlo method.  The rules that
// define that particular patience are listed in detail with the
// comments of the file "table.hh"; files other than "table.*"
// and "solve.cxx" contain definitions and procedures that are
// basically independent from the particular problem to be
// analyzed.
//
// -------------------------------------------------------------

#include <iostream>             // C++ Standard Library headers

#include <cstdlib>              // C Standard Library headers
#include <cstring>
#include <ctime>
#include <cassert>

extern "C" {                    // Unix/POSIX library headers
  #include <unistd.h>
  #include <getopt.h>
}

#include "sol.hh"               // Implementation specific headers
#include "card.hh"

using namespace std;

// Prototypes for global and local functions

void eventLoop(const unsigned long, const unsigned long);
static char *baseName(char *);
static bool getNum(const char *, unsigned long &);
static void usage(char *);

int main(
  int argc,
  char *argv[]
) {

// Tests for internal coherence

  assert(card::totCards          == card::maxValue * card::numSuits);
  assert(sizeof(card::suitName)  == card::numSuits);
  assert(sizeof(card::valueName) == card::maxValue + 1);

// Establishes a Control-C handler; in case of failure, aborts.

  if (!ctrlc()) return EXIT_FAILURE;

// Gets from the command line parameters the number of tries and
// the printout step; then calls eventLoop.

  unsigned long nTries(1000000);
  unsigned long nStep(50000);
  int c;

  opterr = 0;

  while ((c = getopt(argc, argv, ":t:p:")) != -1) {
    switch (c) {
      case 't':
        if (getNum(optarg, nTries)) usage(argv[0]);
        break;
      case 'p':
        if (getNum(optarg, nStep)) usage(argv[0]);
        break;
      case '?':
        cerr << "Unknown option -" << static_cast<char>(optopt)
             << endl;
        usage(argv[0]);
        break;
      case ':':
        cerr << "Option -" << static_cast<char>(optopt)
             << " requires a numerical argument\n";
        usage(argv[0]);
      default:
        cerr << "Can't happen: " << __FILE__ << ", line "
             << __LINE__ << endl;
        return EXIT_FAILURE;
    }
  }

  if (optind < argc) {
    cerr << "Non-option input \"" << argv[optind++];
    while (optind < argc) cerr << ' ' << argv[optind++];
    cerr << "\" will be ignored\n";
  }

#ifdef DEBUG
  cout << "nTries, nStep = " << nTries << ' ' << nStep << endl;
#endif

  clock_t cc = clock();
  eventLoop(nTries, nStep);
  cc = clock() - cc;
  double delta = cc;
  delta /= CLOCKS_PER_SEC;
  cout << delta << " seconds (elapsed CPU time)\n";
  return EXIT_SUCCESS;
}

static char *baseName(
  char *pc
){

// Strips the (eventual) path informations from argv[0], passed as
// formal parameter; if no file name is given, returns an empty
// string.

  assert(pc != 0);

  char *p = strrchr(pc, '/');
  if (p == 0) return pc;
  return ++p;
}

static bool getNum(
  const char *cArray,
  unsigned long &value
) {
  char *pc;
  unsigned long result = strtoul(cArray, &pc, 0);

  if (*pc == '\0') {
    value = result;
    return false;
  } else {
    return true;
  }
}

static void usage(
  char *progName
) {
  cerr << "Usage: " << baseName(progName)
       << " [-t <nr. of tries>] [-p <printout step>]\n";
  exit(EXIT_SUCCESS);
}
