//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
// $Id: table.hh,v 1.3 2005/01/21 07:15:43 loreti Exp $
//
// The table for the patience to be analyzed.
//
// This program simulates a "patience" game, played with a pack
// of 40 cards divided in 4 suits and, for every of these suits,
// having 10 different values.  The cards are disposed in 16
// stacks, arranged on the table in a 4x4 square array like the
//               following figure: the crosses are for stacks of
//   + - - +     two covered cards plus a third, visible, card
//   - + + -     on the top; the dashes are for stacks of one
//   - + + -     covered card plus a second, visible, card on
//   + - - +     the top.  Of course, the actual positions of
//               stacks on the table is completely irrelevant,
// as long as you  have 8 stacks of the first type and 8 of the
// second.  In addition, there are four initially empty
// foundations, one for each suit; the uncovered card at the top
// of every stacks can be moved, in ascending order, over the
// base of its suit; or, in descending or ascending order, over
// another uncovered card of the same suit laying on the top of
// another stack.  Once a covered card is visible, it is turned;
// once a stack is empty, no card can be moved on its location.
//
// -------------------------------------------------------------

#ifndef TABLE_HH_
#define TABLE_HH_

#include <iostream>
#include <cassert>
#include "card.hh"

using std::ostream;

// Forward declaration for class "cardNode"

class cardNode;

// Definitions of the class "cardStack", that is used to
// represent a stack of cards.
//
// The data members are:
// - "head", the root of a single-linked lists of cards (the
//   upper card, always face-up, is the first one, i.e. *head).
//   The list handling is home-brewed; I tried to use std::list
//   from the C++ Standard Library, but the performances have
//   degraded;
// - "howMany", the number of cards in the linked list;
// - "level", used to flag how good is a move for that card
//   stack; is taken equal to the number of _covered_ cards in
//   the stack --- a rough guess, but sufficient for a neat
//   strategy.
//
// The supplied methods are:
// - default constructor, and destructor.  The copy constructor
//   and the assignment operator are declared private and never
//   defined, to exclude their use;
// - an operator > , needed from table::sort();
// - procedures to add a card to the stack (face up, and before
//   the previous first card); to remove the first card from the
//   stack; and to remove all the cards from the stack;
// - inspectors to retrieve the number of cards in the stack
//   and, optionally, also suit and value of the first card;
// - procedures to print a stack on an output stream.

class cardStack {

  friend class table;

  cardNode * head;
  int howMany;
  int level;

  cardStack(const cardStack &);
  cardStack & operator = (const cardStack &);

public:
  cardStack() : head(0), howMany(0), level(-1) { }
  ~cardStack();
  bool operator > (const cardStack &) const;
  void add(const card &);
  const card & remove();
  void removeAll();
  int nCards() const { return howMany; }
  void look(int &, int &, int &) const;
  void print(ostream &) const;
  friend ostream & operator << (ostream & o, const cardStack & cs)
    { cs.print(o);  return o; }
};

// Definition of the class "cardNode", the generic leaf in a
// single-linked list of cards.  The data members are a constant
// reference to a card and a pointer to the next leaf; the
// constructor takes as arguments a const reference to the card
// to be inserted and the two pointers to the parent and the
// child of the new node.

class cardNode {
  friend class cardStack;

  const card & theCard;
  cardNode * next;
  cardNode(const card &, cardNode *, cardNode *);
  ~cardNode() { }
};

// Definition of the class "table".
//
// The data members are:
// - an array of pointers to the 16 stacks.  Pointers, since
//   these stacks will be sorted: and is way more fast to swap
//   two pointers than to swap two cardStacks;
// - the 4 foundations: an array of 4 unsigned integers, holding
//   the value of the last card, for every suit, moved out of
//   the stacks (that are set to 0 initially);
// - the number of the first non-empty stack (when a stack
//   becomes empty, is swapped with the first non-empty one and
//   moved to the beginning);
//
// The supplied methods are:
// - a constructor from an existing card pack.  The default and
//   the copy constructor, and the assignment operator, are
//   declared private and never defined, to exclude their use.
// - a procedure, "solve" that executes the game;
// - procedures to print the table layout on an output stream;
// - a few private procedures, moveIn, moveOut and tableSort,
//   used by solve.

class table {
  cardStack * pStack[16];
  unsigned int lastValue[4];
  unsigned int firstStack;

  table();
  table(const table &);
  table & operator = (const table &);

  bool moveOut();
  bool moveIn();
#ifndef STLSORT
  void tableSort(cardStack *[], int, int);
#endif

public:
  table(const pack &);
  ~table();
  unsigned int solve();
  void print(ostream &) const;
  friend ostream & operator << (ostream & o, const table & t)
    { t.print(o);  return o; }
};

// Inline methods for the class cardStack.
//
// look() : returns how many cards are in the cardStack and,
// *if* the stack is not empty, suit and value of the first
// card.
// nCards() : returns how many cards are in the cardStack.

inline cardStack::~cardStack() {
  removeAll();
}

inline void cardStack::look (
  int &s,
  int &v,
  int &h
) const {

  h = howMany;

  if (head) {
    const card &c = head->theCard;

    s = c.getSuit();
    v = c.getValue();
  }
}

// Inline methods for the class cardNode.
//
// The constructor for the class "cardNode": inserts "newCard"
// in the linked list, between "father" and "child".

inline cardNode::cardNode (
  const card &newCard,
  cardNode *father,
  cardNode *child
) : theCard(newCard), next(child) {
  if (father) father->next = this;
}

#endif
